// TLM CUDA Kernels
// This file contains kernel methods that execute on the GPU and support TLM execution


/*
    Copyright 2010 Computational Electromagnetics Research Laboratory (CERL)
    University of Victoria, Victoria, B.C. Canada
    http://www.cerl.ece.uvic.ca
    http://www.cerl.ece.uvic.ca/Open_Source_CEM.htm
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef _CPP_INTEGRATION_KERNEL_H_
#define _CPP_INTEGRATION_KERNEL_H_

#define VEL_LIGHT  299792458  // m/sec
#define PI 3.1415926





//__constant__ float __align__(16) refl_coef[128];

/*
__constant__ int meshdimx;
__constant__ int meshdimy;
__constant__ int meshdimz;

__constant__ int thread_blk_x;
__constant__ int thread_blk_y;
__constant__ int thread_blk_z;
*/

#define GET_ADDRESS(voltage_arm,x,y,z)	getAddress(voltage_arm,x,y,z,meshDimX,meshDimY,meshDimZ)


////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
////////////////             New Kernels          
int getAddress(int voltage_arm, int x, int y, int z,int meshDimX,int meshDimY,int meshDimZ);



__kernel void
//mem_speed_test_kernel_11( float* data_dev)
//TEST( __global float* data_dev)
//Iterate_scatter( __global float* data_dev)
SCATTER1( __global float* data_dev)
{ 
	
	//const int dimBlocksPerGrid_x = get_num_groups(0); // # of blocks in x-dir of grid
	//const int dimBlocksPerGrid_y = get_num_groups(1); // # of blocks in y-dir of grid
	//int blockID_x = get_group_id(0); // current block ID
	//int blockID_y = get_group_id(1); // current block ID
	//const int threadsPerBlock_x = get_local_size(0); // block dimensions(threads per x-dir in thread block)
	//unsigned int tid_x = get_local_id(0);
	//unsigned int address_YBlock_start; 

	//Reference:   vXX = data_dev[Start of Y\Z Address ThrdBlock  +   X_ThrdBlock__address + X_address + NodeStride*vXX];
	//       i.e.  vXX = data_dev[(X blks)*(64Thds)*(0-Y*Z Blks)  + (64Thds)*(0-X Blks)    + (0-63 Thds) + (64Thds)*(X*Y*Z Blks)*vXX];	
	float v2 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1];
	float v3 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2];
	float v9 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8];
	float v11= data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*10];
	float v1 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0];
	float v12 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11];		
	barrier(CLK_LOCAL_MEM_FENCE);	
	
	// calculate and write-back  v1 v12
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11] = 0.5* ( v2 - v3 + v9  + v11);// v12
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0] = 0.5* ( v2 + v3 + v9  - v11); // v1
	barrier(CLK_GLOBAL_MEM_FENCE);	

	float v6 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*5]; 
	float v10 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*9];
	barrier(CLK_LOCAL_MEM_FENCE);	
	
	// calculate and write-back  v9 v2
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8] = 0.5* ( v1 - v6 + v10 + v12); // v9
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1] = 0.5* ( v1 + v6 - v10 + v12); // v2
	barrier(CLK_GLOBAL_MEM_FENCE);

	float v5 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4];
	float v7 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6];
	float v4 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3];
	float v8 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7];
	barrier(CLK_LOCAL_MEM_FENCE);	

	// calculate and write-back v8 v4
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7] = 0.5* ( v3 - v5 + v7  + v11);  // v8
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3] = 0.5* ( v3 + v5 - v7  + v11);  // v4
	
	// calculate and write-back v11 v3
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*10] = 0.5* (-v1 + v4 + v8  + v12); // v11
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2] = 0.5* ( v1 + v4 + v8  - v12);  // v3
	// calculate  and write-back v10 v6
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*9] = 0.5* (-v2 + v5 + v7  + v9); // v10
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*5] = 0.5* ( v2 + v5 + v7  - v9); // v6
	// calculate  and write-back  v7 v5
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6] = 0.5* (-v4 + v6 + v8  + v10); // v7
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4] = 0.5* ( v4 + v6 - v8  + v10); // v5
	barrier(CLK_GLOBAL_MEM_FENCE);
	
	
	
}




__kernel void

//Iterate_scatter_interchange(__global float* data_dev, int meshDimX, int meshDimY ,int meshDimZ, __local float* tmp_shrd)
SCATTER2(__global float* data_dev, int meshDimX, int meshDimY ,int meshDimZ, __local float* tmp_shrd)
{ 
//Reference: vXX = data_dev[Start_of_Y_ThrdBlock_address    + (Start_of_X_ThrdBlock_Address +Y)  + X_Thrd_Address + Node_address]; 
//    i.e.    v12 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*(12-1)];
// After this kernel (part2) is executed, data is valid for transfer to CPU. 
	const int OFFSET = 3;

	const int dimBlocksPerGrid_x = get_num_groups(0); // # of blocks in x-dir of grid
	const int dimBlocksPerGrid_y = get_num_groups(1); // # of blocks in y-dir of grid
	int blockID_x = get_group_id(0); // current block ID
	int blockID_y = get_group_id(1); // current block ID
	const int threadsPerBlock_x = get_local_size(0); // block dimensions(threads per x-dir in thread block)
	unsigned int tid_x = get_local_id(0);
	unsigned int address_YBlock_start; 
	
	// Reference: +X=11,10 ; -X=3,6 ; Y+=7,12 ; -Y=5,1 ; +Z=9,8 ; -Z=2,4
	float v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12;
	
	/////////////
	// Copy voltages in the +X direction,
	//    and copy them to shared memory, into one index position back (tmp_shrd[] is temporary storage with OFFSET used to ensure array bounds are not violated)
	// +X dir
	tmp_shrd[get_local_id(0) + OFFSET] =0; // clear shared mem
	barrier(CLK_LOCAL_MEM_FENCE);
	tmp_shrd[get_local_id(0) + OFFSET-1] = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2];
	barrier(CLK_LOCAL_MEM_FENCE);
	v11 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0) == 63){
		v11  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + 64 + get_global_size(0)*get_num_groups(1)*2];
	}
	barrier(CLK_LOCAL_MEM_FENCE);	

	// +X dir
	tmp_shrd[get_local_id(0) + OFFSET] =0; // clear shared mem
	barrier(CLK_LOCAL_MEM_FENCE);	
	tmp_shrd[get_local_id(0) + OFFSET-1]  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*5];
	barrier(CLK_LOCAL_MEM_FENCE);	
	v10 = tmp_shrd[get_local_id(0) + OFFSET];
	if (get_local_id(0)== 63){
		v10  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + 64 + get_global_size(0)*get_num_groups(1)*5];
	}
	barrier(CLK_LOCAL_MEM_FENCE);	
	//
	///////////////////////

	/////////////
	// Copy voltages in the -X direction,
	//    and copy them to shared memory, into one index position forward (tmp_shrd[] is temporary storage with OFFSET used to ensure array bounds are not violated)
	// -X dir
	tmp_shrd[get_local_id(0) +OFFSET] =0; // clear shared mem
	barrier(CLK_LOCAL_MEM_FENCE);	
	tmp_shrd[get_local_id(0)+OFFSET+1]  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*10];
	barrier(CLK_LOCAL_MEM_FENCE);	
	v3 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0)== 0){
		v3  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) -1 + get_global_size(0)*get_num_groups(1)*10];
	}
	barrier(CLK_LOCAL_MEM_FENCE);	

	// -X dir
	tmp_shrd[get_local_id(0)+OFFSET] =0; // clear shared mem
	barrier(CLK_LOCAL_MEM_FENCE);	
	tmp_shrd[get_local_id(0)+OFFSET+1	]  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*9];
	barrier(CLK_LOCAL_MEM_FENCE);	
	v6 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0)== 0){
		v6  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) -1 + get_global_size(0)*get_num_groups(1)*9];
	}
	barrier(CLK_LOCAL_MEM_FENCE);	
	//
	//////////////////////


	/////////////////////
	// Copy voltages in +Y and -Y direction
	// easier since +/-Y direction is a differnce in get_local_size(0)*(get_num_groups(0))  in address
	//+Y dir
	//if((get_group_id(1)%meshDimY)<(meshDimY-1)){
		v7	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4];
		v12 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0];
	//}
	//-Y dir
	//if((get_group_id(1)%meshDimY)>(0)){
		v5	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6];
		v1	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11];
	//}
	barrier(CLK_LOCAL_MEM_FENCE);	
	//
	/////////////////////

	/////////////////////
	// Copy voltages in +Z and -Z direction
	// easier since +/-Z direction is a differnce in get_local_size(0)*(get_num_groups(0)*meshDimY)  in address
	//+Z dir
	//if((get_group_id(1)/meshDimY)<(meshDimZ-1)){
		v9	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1];
		v8	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3];
	//}
	//-Zdir
	//if((get_group_id(1)/meshDimY)>(0)){
		v2 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8];
		v4 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7];
	//}
	barrier(CLK_LOCAL_MEM_FENCE);	
	//
	/////////////////////
		
	float v1_r,v2_r,v3_r,v4_r,v5_r,v6_r,v7_r,v8_r,v9_r,v10_r,v11_r,v12_r;

    v1_r= 0.5 * ( v2 + v3 + v9 - v11 );
    v2_r = 0.5 * ( v1 + v6 - v10 + v12 );
    v3_r = 0.5 * ( v1 + v4 + v8 - v12 );
    v4_r = 0.5 * ( v3 + v5 - v7 + v11 );
    v5_r = 0.5 * ( v4 + v6 - v8 + v10 );
    v6_r = 0.5 * ( v2 + v5 + v7 - v9 );
    v7_r = 0.5 * (-v4 + v6 + v8 + v10 );
    v8_r = 0.5 * ( v3 - v5 + v7 + v11 );
    v9_r = 0.5 * ( v1 - v6 + v10 + v12 );
    v10_r = 0.5 * (-v2 + v5 + v7 + v9 );
    v11_r = 0.5 * (-v1 + v4 + v8 + v12 );
    v12_r = 0.5 * ( v2 - v3 + v9 + v11 );
	barrier(CLK_LOCAL_MEM_FENCE);	
	
		// Scatter Calculations and saving results back to Global memory directly
	//+X dir	
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + (get_local_id(0)+1) + get_global_size(0)*get_num_groups(1)*2] = v11_r; // v11->v3  +X	
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + (get_local_id(0)+1) + get_global_size(0)*get_num_groups(1)*5] = v10_r;   // v10->v6  +X
	//-X dir
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + (get_local_id(0)-1) + get_global_size(0)*get_num_groups(1)*10] = v3_r;  // v3->v11  -X
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + (get_local_id(0)-1) + get_global_size(0)*get_num_groups(1)*9] = v6_r;   // v6->v10  -X
	
	//+Y dir
	//if((get_group_id(1)%meshDimY)<(meshDimY-1)){
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0] = v12_r; // v12->v1  +Y
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4] = v7_r;  // v7->v5   +Y
	//}
	//-Y dir
	//if((get_group_id(1)%meshDimY)>(0)){
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11] = v1_r;  // v1->v12  -Y
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6] = v5_r;  // v5->v7   -Y
	//}

	//+Z dir
	//if((get_group_id(1)/meshDimY)<(meshDimZ-1)){
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) +  get_global_size(0)*get_num_groups(1)*1] = v9_r;  // v9->v2  +Z
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3] = v8_r;  // v8->v4  +Z
	//}
	
	//-Zdir
	//if((get_group_id(1)/meshDimY)>(0)){
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7] = v4_r;  // v4->v8  -Z
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8] = v2_r;  // v2->v9  -Z
	//}
		barrier(CLK_GLOBAL_MEM_FENCE);	

		

} // end of  mem_speed_test_kernel_11_part2


__kernel void
//mem_speed_test_kernel_11_part3( float* data_dev)
kernel_v2_TLM_3D_Iterate_scatter_interchange_2(__global float* data_dev, int meshDimX, int meshDimY ,int meshDimZ, __local float* tmp_shrd)
{ 
//Reference: vXX = data_dev[Start_of_Y_ThrdBlock_address    + (Start_of_X_ThrdBlock_Address +Y)  + X_Thrd_Address + Node_address]; 
//    i.e.    v12 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*(12-1)];
// After this kernel (part2) is executed, data is valid for transfer to CPU. 
	//__device__ __shared__ float tmp_shrd[70];

	const int OFFSET = 3;

	const int dimBlocksPerGrid_x = get_num_groups(0); // # of blocks in x-dir of grid
	const int dimBlocksPerGrid_y = get_num_groups(1); // # of blocks in y-dir of grid
	int blockID_x = get_group_id(0); // current block ID
	int blockID_y = get_group_id(1); // current block ID
	const int threadsPerBlock_x = get_local_size(0); // block dimensions(threads per x-dir in thread block)
	unsigned int tid_x = get_local_id(0);
	unsigned int address_YBlock_start; 
	
	// Reference: +X=11,10 ; -X=3,6 ; Y+=7,12 ; -Y=5,1 ; +Z=9,8 ; -Z=2,4
	//float v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12;
	
	//float v1_r,v2_r,v3_r,v4_r,v5_r,v6_r,v7_r,v8_r,v9_r,v10_r,v11_r,v12_r;
	
	float v2,v9;
	//-Zdir
	v2 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8];
	//+Z dir
	v9	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1];

	/////////////
	//////////////////////////////
	//v3 -X dir
	// Copy voltages in the -X direction,
	//    and copy them to shared memory, into one index position forward (tmp_shrd[] is temporary storage with OFFSET used to ensure array bounds are not violated)
	float v3;
	tmp_shrd[get_local_id(0) +OFFSET] =0; // clear shared mem
	barrier(CLK_LOCAL_MEM_FENCE);	
	tmp_shrd[get_local_id(0)+OFFSET+1]  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*10];
	barrier(CLK_LOCAL_MEM_FENCE);
	v3 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0)== 0){
		v3  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) -1 + get_global_size(0)*get_num_groups(1)*10];
	}
	barrier(CLK_LOCAL_MEM_FENCE);
	//
	/////////////////



	/////////////////////////////////
	// v11 +X dir
	// Copy voltages in the +X direction,
	//    and copy them to shared memory, into one index position back (tmp_shrd[] is temporary storage with OFFSET used to ensure array bounds are not violated)
	float v11;
	tmp_shrd[get_local_id(0) + OFFSET] =0; // clear shared mem
	barrier(CLK_LOCAL_MEM_FENCE);
	tmp_shrd[get_local_id(0) + OFFSET-1] = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2];
	barrier(CLK_LOCAL_MEM_FENCE);
	v11 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0) == 63){
		v11  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + 64 + get_global_size(0)*get_num_groups(1)*2];
	}
	
	//
	//////////////
	barrier(CLK_LOCAL_MEM_FENCE);
	float v1_r,v12_r;
	v1_r  = 0.5 * ( v2 + v3 + v9 - v11 );
	v12_r = 0.5 * ( v2 - v3 + v9 + v11 );
	
	//------------------------------------------------

	float v5,v7;
	// -Y dir
	v5	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6];
	// +Y dir
	v7	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4];

	barrier(CLK_LOCAL_MEM_FENCE);


	float v4_r,v8_r,v6_r,v10_r;
	v4_r = 0.5 * ( v3 + v5 - v7 + v11 );
	v8_r = 0.5 * ( v3 - v5 + v7 + v11 );
	v6_r = 0.5 * ( v2 + v5 + v7 - v9 );
	v10_r = 0.5 * (-v2 + v5 + v7 + v9 );
	//-------------------------------------------------


	float v1,v12,v6,v10;
	// -Y dir
	v1	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11];
	// +Y dir
	v12 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0];

	////////////////////////
	// -X dir
	// Copy voltages in the -X direction,
	//    and copy them to shared memory, into one index position forward (tmp_shrd[] is temporary storage with OFFSET used to ensure array bounds are not violated)
	tmp_shrd[get_local_id(0)+OFFSET] =0; // clear shared mem
	barrier(CLK_LOCAL_MEM_FENCE);
	tmp_shrd[get_local_id(0)+OFFSET+1	]  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*9];
	barrier(CLK_LOCAL_MEM_FENCE);
	v6 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0)== 0){
		v6  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) -1 + get_global_size(0)*get_num_groups(1)*9];
	}
	barrier(CLK_LOCAL_MEM_FENCE);
	//
	//////////////////////

	/////////////////////////
	//v10 +X dir
	// Copy voltages in the +X direction,
	//    and copy them to shared memory, into one index position back (tmp_shrd[] is temporary storage with OFFSET used to ensure array bounds are not violated)
	tmp_shrd[get_local_id(0) + OFFSET] =0; // clear shared mem
	barrier(CLK_LOCAL_MEM_FENCE);
	tmp_shrd[get_local_id(0) + OFFSET-1]  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*5];
	barrier(CLK_LOCAL_MEM_FENCE);
	v10 = tmp_shrd[get_local_id(0) + OFFSET];
	if (get_local_id(0)== 63){
		v10  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + 64 + get_global_size(0)*get_num_groups(1)*5];
	}
	barrier(CLK_LOCAL_MEM_FENCE);
	//
	///////////////////////

	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + (get_local_id(0)-1) + get_global_size(0)*get_num_groups(1)*9] = v6_r;   // v6->v10  -X
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + (get_local_id(0)+1) + get_global_size(0)*get_num_groups(1)*5] = v10_r;   // v10->v6  +X
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11] = v1_r;  // v1->v12  -Y
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0] = v12_r; // v12->v1  +Y
	
	barrier(CLK_GLOBAL_MEM_FENCE);

	float v2_r,v9_r;
	v2_r = 0.5 * ( v1 + v6 - v10 + v12 );    
    v9_r = 0.5 * ( v1 - v6 + v10 + v12 );

	//-------------------------------------------------

	//-Zdir
	float v4,v8;
	v4 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7];
	//+Z dir
	v8	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3];
	barrier(CLK_LOCAL_MEM_FENCE);

	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8] = v2_r;  // v2->v9  -Z	
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) +  get_global_size(0)*get_num_groups(1)*1] = v9_r;  // v9->v2  +Z
	
	float v3_r,v11_r,v5_r,v7_r;
    v3_r = 0.5 * ( v1 + v4 + v8 - v12 );
    v11_r = 0.5 * (-v1 + v4 + v8 + v12 );
    v5_r = 0.5 * ( v4 + v6 - v8 + v10 );
    v7_r = 0.5 * (-v4 + v6 + v8 + v10 );

	barrier(CLK_LOCAL_MEM_FENCE);
	
	



	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7] = v4_r;  // v4->v8  -Z
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3] = v8_r;  // v8->v4  +Z

	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + (get_local_id(0)-1) + get_global_size(0)*get_num_groups(1)*10] = v3_r;  // v3->v11  -X	
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + (get_local_id(0)+1) + get_global_size(0)*get_num_groups(1)*2] = v11_r; // v11->v3  +X	
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6] = v5_r;  // v5->v7   -Y	
	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4] = v7_r;  // v7->v5   +Y

	
	barrier(CLK_GLOBAL_MEM_FENCE);


} // end of  mem_speed_test_kernel_11_part3


__kernel void
//mem_speed_test_kernel_11_InterchangeOnly( float* data_dev)
kernel_v2_TLM_3D_Iterate_InterchangeOnly( __global float* data_dev,  int meshDimX, int meshDimY ,int meshDimZ, __local float* tmp_shrd)
{ 
//Reference: vXX = data_dev[Start_of_Y_ThrdBlock_address    + (Start_of_X_ThrdBlock_Address +Y)  + X_Thrd_Address + Node_address]; 
//    i.e.    v12 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*(12-1)];
// After this kernel (part2) is executed, data is valid for transfer to CPU. 
	const int OFFSET = 3;

	const int dimBlocksPerGrid_x = get_num_groups(0); // # of blocks in x-dir of grid
	const int dimBlocksPerGrid_y = get_num_groups(1); // # of blocks in y-dir of grid
	int blockID_x = get_group_id(0); // current block ID
	int blockID_y = get_group_id(1); // current block ID
	const int threadsPerBlock_x = get_local_size(0); // block dimensions(threads per x-dir in thread block)
	unsigned int tid_x = get_local_id(0);
	unsigned int address_YBlock_start; 
	
	// Reference: +X=11,10 ; -X=3,6 ; Y+=7,12 ; -Y=5,1 ; +Z=9,8 ; -Z=2,4
	float v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12;
	
	/////////////
	// Copy voltages in the +X direction,
	//    and copy them to shared memory, into one index position back (tmp_shrd[] is temporary storage with OFFSET used to ensure array bounds are not violated)
	// +X dir
	tmp_shrd[get_local_id(0)+OFFSET-1] = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*10];
	barrier(CLK_LOCAL_MEM_FENCE);
	v11 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0)== 63){
		v11  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + 64 + get_global_size(0)*get_num_groups(1)*10];
	}
	barrier(CLK_LOCAL_MEM_FENCE);

	// +X dir
	tmp_shrd[get_local_id(0)+OFFSET-1]  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*9];
	barrier(CLK_LOCAL_MEM_FENCE);
	v10 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0)== 63){
		v10  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + 64 + get_global_size(0)*get_num_groups(1)*9];
	}
	barrier(CLK_LOCAL_MEM_FENCE);
	//
	///////////////////////

	/*
	/////////////
	// Copy voltages in the -X direction,
	//    and copy them to shared memory, into one index position forward (tmp_shrd[] is temporary storage with OFFSET used to ensure array bounds are not violated)
	// -X dir
	tmp_shrd[get_local_id(0)+OFFSET+1]  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2];
	barrier(CLK_LOCAL_MEM_FENCE);
	v3 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0)== 0){
		v3  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) -1 + get_global_size(0)*get_num_groups(1)*2];
	}

	// -X dir
	tmp_shrd[get_local_id(0)+OFFSET+1]  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*5];
	barrier(CLK_LOCAL_MEM_FENCE);
	v6 = tmp_shrd[get_local_id(0)+OFFSET];
	if (get_local_id(0)== 0){
		v6  = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) -1 + get_global_size(0)*get_num_groups(1)*5];
	}
	//
	//////////////////////
	*/

	/////////////////////
	// Copy voltages in +Y and -Y direction
	// easier since +/-Y direction is a differnce in get_local_size(0)*(get_num_groups(0))  in address
	//+Y dir
	if((get_group_id(1)%meshDimY)<(meshDimY-1)){
		v7	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6];
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6] = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6];
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6] = v7;
		
		v12 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11];
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11] = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11];
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11] = v12;
		
	}
	barrier(CLK_GLOBAL_MEM_FENCE);
	/*
	//-Y dir
	if((get_group_id(1)%meshDimY)>(0)){
		v5	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4];
		v1	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0];
	}
	//
	/////////////////////
	*/

	/////////////////////
	// Copy voltages in +Z and -Z direction
	// easier since +/-Z direction is a differnce in get_local_size(0)*(get_num_groups(0)*meshDimY)  in address
	//+Z dir
	if((get_group_id(1)/meshDimY)<(meshDimZ-1)){
		v9	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8];
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8] = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8];
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8] = v9;


		v8	 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7];
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)+get_num_groups(0)) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7] = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7];
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7] = v8;

	}
	barrier(CLK_GLOBAL_MEM_FENCE);
	/*
	//-Zdir
	if((get_group_id(1)/meshDimY)>(0)){
		v2 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1];
		v4 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*(get_group_id(0)-get_num_groups(0)*meshDimY) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3];
	}
	*/
	

	//
	/////////////////////


	// Save data back to memory
	/*
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*11] = v12;
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0] = v1;

		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*8] = v9;
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1] = v2;

		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*7] = v8;
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3] = v4;

		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*10] = v11;
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2] = v3;

		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*9] = v10;
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*5] = v6;

		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*6] = v7;
		data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4] = v5;
		__syncthreads();
		*/


} // end of  mem_speed_test_kernel_11_part2


__kernel void
//mem_speed_test_kernel_11_Boundary( float* data_dev,int Xmin,int Xmax, int Ymin, int Ymax, int Zmin,int Zmax, float reflection_coef )
kernel_v2_TLM_3D_Boundary(	__global float* data_dev,
							int Xmin,
							int Xmax,
							int Ymin,
							int Ymax, 
							int Zmin,
							int Zmax, 
							float reflection_coef ,
							int meshDimX,
							int meshDimY,
							int meshDimZ
						){ 
	const int dimBlocksPerGrid_x = get_num_groups(0); // # of blocks in x-dir of grid
	const int dimBlocksPerGrid_y = get_num_groups(1); // # of blocks in y-dir of grid
	const int threadsPerBlock_x = get_local_size(0); // block dimensions(threads per x-dir in thread block)
	const int threadsPerBlock_y = get_local_size(1); // block dimensions(threads per y-dir in thread block)
	unsigned blockID_x = get_group_id(0); // current block ID
	unsigned blockID_y = get_group_id(1); // current block ID
	unsigned int tid_x = get_local_id(0);
	unsigned int tid_y = get_local_id(1);
	float temp1,temp2;

	/////// TLM SCN Node Structure ////////
	//            z7,x12                 //
	//  y3,z6<-      /y4,x2   ->y11,z10  // key: up=+y,right=+x, out-of-page=+z
	//         y8,x9/                    // [polz dir][node#] i.e. 'y3'
	//            z5,x1                  //
	///////////////////////////////////////
	
	

	if (Zmin == Zmax){ // XY Plane
		int X_coord = Xmin + threadsPerBlock_x*blockID_x + tid_x;
		int Y_coord = Ymin + threadsPerBlock_y*blockID_y + tid_y;
		int Z_coord = Zmin ;

		if ((X_coord <= Xmax) && ( Y_coord <= Ymax) ){
			temp1 = data_dev[GET_ADDRESS(8,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(8,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(4,X_coord,Y_coord,Z_coord+1)];
			data_dev[GET_ADDRESS(4,X_coord,Y_coord,Z_coord+1)] = reflection_coef *temp1;

			temp2 = data_dev[GET_ADDRESS(9,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(9,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(2,X_coord,Y_coord,Z_coord+1)];
			data_dev[GET_ADDRESS(2,X_coord,Y_coord,Z_coord+1)] = reflection_coef *temp2;

		} // end of if
	} else if(Ymin == Ymax){ // XZ plane
		int X_coord = Xmin + threadsPerBlock_x*blockID_x + tid_x;
		int Y_coord = Ymin ;
		int Z_coord = Zmin + threadsPerBlock_y*blockID_y + tid_y;

		if ((X_coord <= Xmax) && ( Z_coord <= Zmax)){
			temp1 = data_dev[GET_ADDRESS(12,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(12,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(1,X_coord,Y_coord+1,Z_coord)];
			data_dev[GET_ADDRESS(1,X_coord,Y_coord+1,Z_coord)] = reflection_coef *temp1;

			temp2 = data_dev[GET_ADDRESS(7,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(7,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(5,X_coord,Y_coord+1,Z_coord)];
			data_dev[GET_ADDRESS(5,X_coord,Y_coord+1,Z_coord)] = reflection_coef *temp2;
		} // end of if

	} else if(Xmin == Xmax){ // YZ plane
		int X_coord = Xmin ;
		int Y_coord = Ymin + threadsPerBlock_x*blockID_x + tid_x;
		int Z_coord = Zmin + threadsPerBlock_y*blockID_y + tid_y;
		if (( Y_coord <= Ymax)&&( Z_coord <= Zmax)){
			temp1 = data_dev[GET_ADDRESS(11,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(11,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(3,X_coord+1,Y_coord,Z_coord)];
			data_dev[GET_ADDRESS(3,X_coord+1,Y_coord,Z_coord)] = reflection_coef *temp1;

			temp2 = data_dev[GET_ADDRESS(10,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(10,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(6,X_coord+1,Y_coord,Z_coord)];
			data_dev[GET_ADDRESS(6,X_coord+1,Y_coord,Z_coord)] = reflection_coef *temp2;
		} // end of if
	} // end of else if
	barrier(CLK_GLOBAL_MEM_FENCE);


} // end of kernel



__kernel void
kernel_v2_TLM_3D_Boundary_2(__global float* data_dev,
							int Xmin,
							int Xmax, 
							int Ymin, 
							int Ymax, 
							int Zmin,
							int Zmax, 
							float reflection_coef,
							int meshDimX,
							int meshDimY,
							int meshDimZ
						){ 
	const int dimBlocksPerGrid_x = get_num_groups(0); // # of blocks in x-dir of grid
	const int dimBlocksPerGrid_y = get_num_groups(1); // # of blocks in y-dir of grid
	const int threadsPerBlock_x = get_local_size(0); // block dimensions(threads per x-dir in thread block)
	const int threadsPerBlock_y = get_local_size(1); // block dimensions(threads per y-dir in thread block)
	unsigned blockID_x = get_group_id(0); // current block ID
	unsigned blockID_y = get_group_id(1); // current block ID
	unsigned int tid_x = get_local_id(0);
	unsigned int tid_y = get_local_id(1);
	float temp1,temp2;

	/////// TLM SCN Node Structure ////////
	//            z7,x12                 //
	//  y3,z6<-      /y4,x2   ->y11,z10  // key: up=+y,right=+x, out-of-page=+z
	//         y8,x9/                    // [polz dir][node#] i.e. 'y3'
	//            z5,x1                  //
	///////////////////////////////////////
	
	if (Zmin == Zmax){ // XY Plane
		int X_coord = (Xmin/16)*16 + threadsPerBlock_x*blockID_x + tid_x;
		int Y_coord = (Ymin/16)*16 + threadsPerBlock_y*blockID_y + tid_y;
		int Z_coord = Zmin ;

		if ((X_coord <= Xmax) && ( Y_coord <= Ymax) &&(X_coord >= Xmin) && ( Y_coord >= Ymin) ){
			temp1 = data_dev[GET_ADDRESS(8,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(8,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(4,X_coord,Y_coord,Z_coord+1)];
			data_dev[GET_ADDRESS(4,X_coord,Y_coord,Z_coord+1)] = reflection_coef *temp1;

			temp2 = data_dev[GET_ADDRESS(9,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(9,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(2,X_coord,Y_coord,Z_coord+1)];
			data_dev[GET_ADDRESS(2,X_coord,Y_coord,Z_coord+1)] = reflection_coef *temp2;

		} // end of if
	} else if(Ymin == Ymax){ // XZ plane
		int X_coord = (Xmin/16)*16 + threadsPerBlock_x*blockID_x + tid_x;
		int Y_coord = Ymin ;
		int Z_coord = (Zmin/16)*16 + threadsPerBlock_y*blockID_y + tid_y;

		if ((X_coord <= Xmax) && ( Z_coord <= Zmax) && (X_coord >= Xmin) && ( Z_coord >= Zmin)){
			temp1 = data_dev[GET_ADDRESS(12,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(12,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(1,X_coord,Y_coord+1,Z_coord)];
			data_dev[GET_ADDRESS(1,X_coord,Y_coord+1,Z_coord)] = reflection_coef *temp1;

			temp2 = data_dev[GET_ADDRESS(7,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(7,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(5,X_coord,Y_coord+1,Z_coord)];
			data_dev[GET_ADDRESS(5,X_coord,Y_coord+1,Z_coord)] = reflection_coef *temp2;
		} // end of if

	} else if(Xmin == Xmax){ // YZ plane
		int X_coord = Xmin ;
		int Y_coord = (Ymin/16)*16 + threadsPerBlock_x*blockID_x + tid_x;
		int Z_coord = (Zmin/16)*16 + threadsPerBlock_y*blockID_y + tid_y;
		if (( Y_coord <= Ymax) && ( Z_coord <= Zmax) && ( Y_coord >= Ymin)&&( Z_coord >= Zmin)){
			temp1 = data_dev[GET_ADDRESS(11,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(11,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(3,X_coord+1,Y_coord,Z_coord)];
			data_dev[GET_ADDRESS(3,X_coord+1,Y_coord,Z_coord)] = reflection_coef *temp1;

			temp2 = data_dev[GET_ADDRESS(10,X_coord,Y_coord,Z_coord )];
			data_dev[GET_ADDRESS(10,X_coord,Y_coord,Z_coord )] = reflection_coef * data_dev[GET_ADDRESS(6,X_coord+1,Y_coord,Z_coord)];
			data_dev[GET_ADDRESS(6,X_coord+1,Y_coord,Z_coord)] = reflection_coef *temp2;
		} // end of if
	} // end of else if
	barrier(CLK_GLOBAL_MEM_FENCE);


} // end of kernel


__kernel void
//mem_speed_test_kernel_11_Plane_Excitation( float* data_dev,int Xmin,int Xmax, int Ymin, int Ymax, int Zmin, int Zmax, int voltage_arm, float excitation, int expanded_node_state )// propagation in +/- Z dir
kernel_v2_TLM_3D_Plane_Excitation(	__global float* data_dev,
									int Xmin,
									int Xmax, 
									int Ymin, 
									int Ymax, 
									int Zmin, 
									int Zmax, 
									int voltage_arm, 
									float excitation, 
									int expanded_node_state,
									int meshDimX,
									int meshDimY,
									int meshDimZ					
									)// propagation in +/- Z dir
{ 

	const int dimBlocksPerGrid_x = get_num_groups(0); // # of blocks in x-dir of grid
	const int dimBlocksPerGrid_y = get_num_groups(1); // # of blocks in y-dir of grid
	const int threadsPerBlock_x = get_local_size(0); // block dimensions(threads per x-dir in thread block)
	const int threadsPerBlock_y = get_local_size(1); // block dimensions(threads per y-dir in thread block)
	unsigned blockID_x = get_group_id(0); // current block ID
	unsigned blockID_y = get_group_id(1); // current block ID
	unsigned int tid_x = get_local_id(0);
	unsigned int tid_y = get_local_id(1);

	if (Zmin == Zmax){ // XY plane Z propagation v2,v4 or v8,v9
		int X_coord = Xmin + threadsPerBlock_x*blockID_x + tid_x;
		int Y_coord = Ymin + threadsPerBlock_y*blockID_y + tid_y;
		int Z_coord = Zmin;
		if ((X_coord <= Xmax) && ( Y_coord <= Ymax) &&(Z_coord<=Zmax) ) {
			data_dev[GET_ADDRESS(voltage_arm,X_coord,Y_coord,(Z_coord + expanded_node_state) )] = excitation;
		}// end of if

	} else if (Ymin == Ymax){ // XZ Plane Y propagation v1,v5 or v12,v7
		int X_coord = Xmin + threadsPerBlock_x*blockID_x + tid_x;
		int Y_coord = Ymin ;
		int Z_coord = Zmin + threadsPerBlock_y*blockID_y + tid_y;
		if ((X_coord <= Xmax) && ( Y_coord <= Ymax) &&(Z_coord<=Zmax) ) {
			data_dev[GET_ADDRESS(voltage_arm,X_coord,(Y_coord + expanded_node_state),Z_coord  )] = excitation;
		}// end of if

	}else if (Xmin == Xmax){ // YZ Plane X propagation v3,v6 or v11,10
		int X_coord = Xmin ;
		int Y_coord = Ymin + threadsPerBlock_x*blockID_x + tid_x;
		int Z_coord = Zmin + threadsPerBlock_y*blockID_y + tid_y;
		if ((X_coord <= Xmax) && ( Y_coord <= Ymax) &&(Z_coord<=Zmax) ) {
			data_dev[GET_ADDRESS(voltage_arm,(X_coord + expanded_node_state),Y_coord,Z_coord)] = excitation;
		}// end of if
	}// end of else-if

} // end of kernel



__kernel void
//mem_speed_test_kernel_11_Impulse_Excitation( float* data_dev,int x,int y, int z, float excitation, int expanded_node_state )
kernel_v2_TLM_3D_Plane_Impulse_Excitation(	__global float* data_dev,
											int x,
											int y, 
											int z, 
											float excitation, 
											int expanded_node_state ,
											int meshDimX,
											int meshDimY,
											int meshDimZ
										){
	if (expanded_node_state==0){
		data_dev[GET_ADDRESS(1,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(2,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(3,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(4,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(5,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(6,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(7,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(8,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(9,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(10,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(11,x,y,z)] += excitation;
		data_dev[GET_ADDRESS(12,x,y,z)] += excitation;
	} else {
		data_dev[GET_ADDRESS(1,x,y-1,z)] += excitation;
		data_dev[GET_ADDRESS(2,x,y,z-1)] += excitation;
		data_dev[GET_ADDRESS(3,x-1,y,z)] += excitation;
		data_dev[GET_ADDRESS(4,x,y,z-1)] += excitation;
		data_dev[GET_ADDRESS(5,x,y-1,z)] += excitation;
		data_dev[GET_ADDRESS(6,x-1,y,z)] += excitation;
		data_dev[GET_ADDRESS(7,x,y+1,z)] += excitation;
		data_dev[GET_ADDRESS(8,x,y,z+1)] += excitation;
		data_dev[GET_ADDRESS(9,x,y,z+1)] += excitation;
		data_dev[GET_ADDRESS(10,x+1,y,z)] += excitation;
		data_dev[GET_ADDRESS(11,x+1,y,z)] += excitation;
		data_dev[GET_ADDRESS(12,x,y+1,z)] += excitation;
	}

	barrier(CLK_GLOBAL_MEM_FENCE);

}// end of kernel

__kernel void
//kernel_TLM_3D_Plane_HalfSine_Excitation_new(float* data_dev,int dir_polrz, int u_min,int u_max,int v_min,int v_max,int w_plane, float excitation, int expanded_node_state){
kernel_v2_TLM_3D_Plane_HalfSine_Excitation(	__global float* data_dev,
											int dir_polrz, 
											int u_min,
											int u_max,
											int v_min,
											int v_max,
											int w_plane, 
											float excitation, 
											int expanded_node_state,
											int meshDimX,
											int meshDimY,
											int meshDimZ
			){
	// xmin,xmax,ymin,ymax, zmin,zmax, define a sheet excitation 
	// The halfsin excitation is a sheet excitation propagating in the '+w' or '-w' direction.
	//  It is esentially a fundamental mode TE10 propagating in the 'w' direction. Half sine is along 'u' direction. 
	//   Polerization is along 'v'direction. The u,v,w translate to the x,y,z convention by the using 'dir_polrz'. 
	//   dir_polz = voltage arm: use the TLM SCN node model to find direction of propagation and polerization
	//      i.e. dir_polz = 1 (v1)  then polerization is in x direction and propatation is in +Y direction. 
	// note: that v1 is actually in the -Y position on the SCN node structure but we say that direction of propagation is +veY
	//       since exciting this voltage arm will launch a +ve Y impulse. 
	
	// Setup must be: 

	//Convention: u=span of half sine wave distribution
	//			  v = polarization direction
	//            w = propogation direction
	//ALL = 0, XnegYpol = 11, XposYpol = 3, XnegZpol = 10, XposZpol = 6, 
	//		   YnegXpol = 12, YposXpol = 1, YnegZpol =  7, YposZpol = 5, 
	//		   ZnegXpol =  9, ZposXpol = 2, ZnegYpol =  8, ZposYpol = 4
	// examples: XposYpol  = propogation in the postive  X direction with poloarization in the Y direction
	//           ZnegXpol  = propogation in the negative Z direction with poloarization in the X direction

	// Block index
    int bu = get_group_id(0);
    int bv = get_group_id(1);

	int block_u_dim = get_local_size(0); // block dimensions (ie. 20 threads/block)
	int block_v_dim = get_local_size(1); 

	int tu = get_local_id(0) ;  // used to map to shared memory
    int tv = get_local_id(1) ;

	int u_idx = u_min + block_u_dim*bu + tu;
	int v_idx = v_min + block_v_dim*bv + tv;			
	
	float ex_temp_u = 0;
	float u_width = (float)(u_max - u_min + 1);
	//float u_width = (float)(u_max - u_min);
	
	float u_temp = (float)(u_idx - u_min + 0.5);
	//float u_temp = (float)(u_idx - u_min);
	ex_temp_u = sin((float)(PI*u_temp/u_width))*excitation;
	
	
	//cube_point_add = cube_address(v_idx,u_idx, w_plane);
		
	if ( (u_idx <= u_max) && (v_idx <= v_max)){
		switch(dir_polrz){
			case 0 : 
				data_dev[GET_ADDRESS(1,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(2,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(3,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(4,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(5,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(6,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(7,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(8,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(9,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(10,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(11,u_idx,v_idx,w_plane)] += ex_temp_u;
				data_dev[GET_ADDRESS(12,u_idx,v_idx,w_plane)] += ex_temp_u;
				break;
			case 1: // Y prop X polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(1,v_idx,w_plane,u_idx )] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(12,v_idx,w_plane-1,u_idx )] += ex_temp_u;
				break;
			case 2: // Z prop X polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(2,v_idx,u_idx,w_plane )] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(9,v_idx,u_idx,w_plane-1 )] += ex_temp_u;
				break;
			case 3: // X prop Y polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(3,w_plane, v_idx, u_idx)] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(11,w_plane-1, v_idx, u_idx)] += ex_temp_u;
				break;
			case 4: // Z prop Y polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(4,u_idx, v_idx ,w_plane)] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(8,u_idx, v_idx ,w_plane-1)] += ex_temp_u;
				break;
			case 5: // Y prop Z polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(5,u_idx, w_plane, v_idx )] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(7,u_idx, w_plane-1, v_idx )] += ex_temp_u;
				break;
			case 6: // X prop Z polz
				if (expanded_node_state==0){
					data_dev[GET_ADDRESS(6,w_plane, u_idx,  v_idx )] += ex_temp_u;
				} else{
					data_dev[GET_ADDRESS(10,w_plane-1, u_idx,  v_idx )] += ex_temp_u;}
				break;
			case 7: // -Y prop Z polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(7,w_plane, u_idx,  v_idx )] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(5,w_plane+1, u_idx,  v_idx )] += ex_temp_u;
				break;
			case 8: // -Z prop Y polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(8,u_idx, v_idx, w_plane)] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(4,u_idx, v_idx, w_plane+1)] += ex_temp_u;
				break;
			case 9: // -Z prop X polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(9,v_idx ,u_idx,  w_plane)] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(2,v_idx ,u_idx,  w_plane+1)] += ex_temp_u;
				break;
			case 10: // -X prop Z polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(10,w_plane  , u_idx ,v_idx)] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(6,w_plane+1,u_idx ,v_idx)] += ex_temp_u;
				break;
			case 11: // -X prop Y polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(11, w_plane, v_idx, u_idx)] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(3, w_plane+1, v_idx, u_idx)] += ex_temp_u;
				break;
			case 12: // -Y prop X polz
				if (expanded_node_state==0)
					data_dev[GET_ADDRESS(12, v_idx, w_plane ,u_idx)] += ex_temp_u;
				else
					data_dev[GET_ADDRESS(1, v_idx, w_plane+1,u_idx)] += ex_temp_u;
				break;
			default:
				break;
		}// end of case

	}/// end of if
	barrier(CLK_GLOBAL_MEM_FENCE);


}


__kernel void
//mem_speed_test_kernel_11_Sample( float* data_dev ,float* GPU_Samples_ptr,int sample_index,int polerize_dir,  int x , int y , int z, int expanded_node_state  ){
kernel_v2_TLM_3D_Sample(__global float* data_dev,
						__global float* GPU_Samples_ptr,
						int sample_index,
						int polerize_dir,  
						int x , 
						int y , 
						int z, 
						int expanded_node_state,
						int meshDimX,
						int meshDimY,
						int meshDimZ
){
	float sample_result ;
	if (polerize_dir == 1){		   // X-polerization
		sample_result = ( data_dev[GET_ADDRESS(1,x, y + expanded_node_state, z)] + data_dev[GET_ADDRESS(2,x, y, z+ expanded_node_state)] + data_dev[GET_ADDRESS(12,x, y - expanded_node_state, z)] + data_dev[GET_ADDRESS(9,x, y, z - expanded_node_state)] )/4.0;
	} else if (polerize_dir == 2){ // Y-polerization
		sample_result = ( data_dev[GET_ADDRESS(3,x + expanded_node_state, y, z)] + data_dev[GET_ADDRESS(4,x, y, z + expanded_node_state)] + data_dev[GET_ADDRESS(11,x - expanded_node_state, y, z)] + data_dev[GET_ADDRESS(8,x, y, z - expanded_node_state)] )/4.0;
	} else if (polerize_dir == 3){ // Z-polerization
		sample_result = ( data_dev[GET_ADDRESS(5,x, y + expanded_node_state, z)] + data_dev[GET_ADDRESS(6,x + expanded_node_state, y, z)] + data_dev[GET_ADDRESS(7,x, y - expanded_node_state, z)] + data_dev[GET_ADDRESS(10,x - expanded_node_state, y, z)] )/4.0;
	//} else {
	//	sample_result = (	data_dev[GET_ADDRESS(1,x, y, z)] + data_dev[GET_ADDRESS(4,x, y, z)] + data_dev[GET_ADDRESS(7,x, y, z)] + data_dev[GET_ADDRESS(10,x, y, z)] +
	//						data_dev[GET_ADDRESS(2,x, y, z)] + data_dev[GET_ADDRESS(5,x, y, z)] + data_dev[GET_ADDRESS(8,x, y, z)] + data_dev[GET_ADDRESS(11,x, y, z)] +
	//						data_dev[GET_ADDRESS(3,x, y, z)] + data_dev[GET_ADDRESS(6,x, y, z)] + data_dev[GET_ADDRESS(9,x, y, z)] + data_dev[GET_ADDRESS(12,x, y, z)]   )/4.0;
	}
	GPU_Samples_ptr[sample_index] = sample_result;
	barrier(CLK_GLOBAL_MEM_FENCE);
}// end of kernel

int getAddress(int voltage_arm, int x, int y, int z, int meshDimX, int meshDimY, int meshDimZ){
	return meshDimZ*meshDimY*meshDimX*(voltage_arm - 1) + meshDimY*meshDimX*z + meshDimX*y + x;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//// 2D TLM code
//
//__kernel void
//TLM_2D_Scatter( __global float* data_dev)
//{ 
//
//	//Reference:   vXX = data_dev[Start of Y Address ThrdBlock  +   X_ThrdBlock__address + X_address + NodeStride*vXX];
//	float v0 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0];
//	float v1 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1];
//	float v2 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2];
//	float v3 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3];
//	float v4 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4];
//	barrier(CLK_LOCAL_MEM_FENCE);	
//	
//	float rdcpy = 0;
//	float a = (v1 + v2 + v3 + v4 + v0*rdcpy)*2/(rdcpy+4);
//	barrier(CLK_GLOBAL_MEM_FENCE);	
//
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0] = a - v0;
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1] = a - v1;
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2] = a - v2;
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3] = a - v3;
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4] = a - v4;
//	barrier(CLK_GLOBAL_MEM_FENCE);	
//
//}
//
//__kernel void
//TLM_2D_Scatter_Impulse(__global float* data_dev, int meshDimX, int meshDimY ,int meshDimZ, __local float* tmp_shrd)
//{ 
//	//Reference:   vXX = data_dev[Start of Y Address ThrdBlock  +   X_ThrdBlock__address + X_address + NodeStride*vXX];
//	float v0 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0];
//	float v1 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1];
//	float v2 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2];
//	float v3 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3];
//	float v4 = data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4];
//	barrier(CLK_LOCAL_MEM_FENCE);	
//	
//	float rdcpy = 0;
//	float a = (v1 + v2 + v3 + v4 + v0*rdcpy)*2/(rdcpy+4);
//	barrier(CLK_GLOBAL_MEM_FENCE);	
//
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*0] = a - v0;
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*1] = a - v1;
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*2] = a - v2;
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*3] = a - v3;
//	data_dev[get_global_size(0)*get_group_id(1) +  get_local_size(0)*get_group_id(0) + get_local_id(0) + get_global_size(0)*get_num_groups(1)*4] = a - v4;
//	barrier(CLK_GLOBAL_MEM_FENCE);	
//
//}



#endif // #ifndef _CPP_INTEGRATION_KERNEL_H_